
/*
 * $XConsortium$
 *
 * Copyright 1989 Massachusetts Institute of Technology
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of M.I.T. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  M.I.T. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * M.I.T. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL M.I.T.
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Authors:  Davor Matic, MIT X Consortium
 *           Thomas Wu, MIT Fishbowl
 */

#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/XWDFile.h>
#include "xwd.h"

#define MAXCMD 128

/*
 * Read in the header, window name, colormap, and pixmap from the input file.
 */

ReadXwdFile(file_name, image, colors, ncolors)
     String file_name;
     XImage **image;
     XColor **colors;
     int *ncolors;
{
  Bool pipeline = False;
  FILE *file;
  char shell_cmd[MAXCMD];
  int status;

  if (strlen(file_name) > 1 &&
      !strcmp(file_name + strlen(file_name) - 2, ".Z")) {
    (void) sprintf(shell_cmd, "zcat %s", file_name);
    if((file = popen(shell_cmd, "r")) == NULL)
      return RX_ZCAT_ERROR;
    pipeline = True;
  }
  else if (strcmp(file_name, "-")) {
    /*
     * Open the output file.
     */
    file = fopen(file_name, "r");
    if (!file)
      return RX_FILE_NOT_FOUND;
  }
  else {
    file = stdin;
  }
  status = ReadXwdDescriptor(file, image, colors, ncolors);
  if(pipeline)
    (void) pclose(file);
  else if(file != stdin)
    (void) fclose(file);
  return status;
}

unsigned long DataSize();

ReadXwdDescriptor(file, image, colors, ncolors)
    FILE *file;
    XImage **image;
    XColor **colors;
    int *ncolors;
{
    XWDFileHeader header;
    String win_name;
    int i;
    register char *buffer;
    unsigned buffer_size, win_name_size;
    unsigned long swaptest = 1;
    int status = RX_SUCCESS;

    /*
     * Read in header information.
     */
    if(fread((char *)&header, sizeof(header), 1, file) != 1)
	return RX_BAD_HEADER;

    if (*(char *) &swaptest)
	_swaplong((char *) &header, sizeof(header));

    /*
     * Check if the dump file is in the proper format.
     */
    if (header.file_version != XWD_FILE_VERSION)
	return RX_NOT_XWD_FILE;

    if (header.header_size < sizeof(header))
	return RX_BAD_HEADER;

    /*
     * Malloc window name.
     */
    win_name_size = (header.header_size - sizeof(header));
    win_name = (String) XtMalloc((unsigned) win_name_size * sizeof(char));

    /*
     * Read in window name.
     */
    if(fread(win_name, sizeof(char), (int)win_name_size, file)!= win_name_size)
	status = RX_BAD_WIN_NAME;

    /*
     * Malloc the colormap buffer.
     */
    if(*ncolors = header.ncolors) {
	*colors = (XColor *) XtMalloc((unsigned) *ncolors * sizeof(XColor));
	if(fread((char *) *colors, sizeof(XColor), *ncolors, file) != *ncolors)
	    return RX_BAD_COLORTABLE;
    }
    
    if (*(char *) &swaptest) {
	for (i = 0; i < *ncolors; i++) {
	    _swaplong((char *) &(*colors)[i].pixel, sizeof(long));
	    _swapshort((char *) &(*colors)[i].red, 3 * sizeof(short));
	}
    }
    
    /*
     * Malloc the pixel buffer.
     */
    buffer_size = DataSize((int)header.bytes_per_line,
			   (int)header.pixmap_height,
			   (int)header.pixmap_depth, 
			   (int)header.pixmap_format);
    buffer = XtMalloc(buffer_size);

    /*
     * Malloc the image.
     */
    *image = (XImage *) XtMalloc(sizeof(XImage));
    
    /*
     * Initialize image.
     */
    (*image)->data = (char *) buffer;
    (*image)->width = (int) header.pixmap_width;
    (*image)->height = (int) header.pixmap_height;
    (*image)->xoffset = (int) header.xoffset;
    (*image)->format = (int) header.pixmap_format;
    (*image)->byte_order = (int) header.byte_order;
    (*image)->bitmap_unit = (int) header.bitmap_unit;
    (*image)->bitmap_bit_order = (int) header.bitmap_bit_order;
    (*image)->bitmap_pad = (int) header.bitmap_pad;
    (*image)->depth = (int) header.pixmap_depth;
    (*image)->bits_per_pixel = (int) header.bits_per_pixel;
    (*image)->bytes_per_line = (int) header.bytes_per_line;
    (*image)->red_mask = header.red_mask;
    (*image)->green_mask = header.green_mask;
    (*image)->blue_mask = header.blue_mask;
    (*image)->obdata = NULL;
    _XInitImageFuncPtrs(*image);

    /****************************&&&&&&&&&&%%%%%%%%%%%%%%%%%*/
    if((*image)->depth == 1)
      (*image)->format = XYBitmap;

    /*
     * Read in the pixmap buffer.
     */
    if(fread((char *) buffer, sizeof(char), (int)buffer_size, file)
       != buffer_size)
	return RX_CORRUPT;

    return status;
}

_swapshort (bp, n)
    register char *bp;
    register unsigned n;
{
    register char c;
    register char *ep = bp + n;

    while (bp < ep) {
	c = *bp;
	*bp = *(bp + 1);
	bp++;
	*bp++ = c;
    }
}

_swaplong (bp, n)
    register char *bp;
    register unsigned n;
{
    register char c;
    register char *ep = bp + n;
    register char *sp;

    while (bp < ep) {
	sp = bp + 3;
	c = *sp;
	*sp = *bp;
	*bp++ = c;
	sp = bp + 1;
	c = *sp;
	*sp = *bp;
	*bp++ = c;
	bp += 2;
    }
}
