
/*
 * $XConsortium$
 *
 * Copyright 1989 Massachusetts Institute of Technology
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of M.I.T. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  M.I.T. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * M.I.T. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL M.I.T.
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Authors:  Davor Matic, MIT X Consortium
 *           Thomas Wu, MIT Fishbowl
 */

#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Label.h>
#include "Scale.h"

#define min(x, y) ((x) < (y) ? (x) : (y))
#define max(x, y) ((x) > (y) ? (x) : (y))

Widget toplevel, scale, shell, label;
Colormap colormap;

static String usage = "[-options ...]\n\
\n\
where options include:\n\
    all standart toolkit options\n\
    -bitmap filename\n\
    -xwd filename\n\
    -gif filename                 file to be displayed\n\
    -inc, +inc                    incorporate defalut colormap\n\
    -install, +install            install into default colormap\n\
    -limit ncolors                limit the number of colors to be used\n\
    -ar aspect ratio              x / y * aspect ratio = 1\n\
    -scale float                  ...\n\
    -p, +p                        proportionsl mode on, off\n\
    -a, +a                        autoscale on, off\n\
";

Usage()
{
    fprintf(stderr, usage);
    exit(1);
}

static struct AppResources {
    Cardinal limit;
    Boolean  incorporate;
    Boolean  install;
} app_resources;

static XtResource resources[] = {
#define offset(field) XtOffset(struct AppResources *, field)
{
    "limit", "Limit", XtRInt, sizeof(int),
    offset(limit), XtRImmediate, (XtPointer) 0},
{
    "incorporate", "Incorporate", XtRBoolean, sizeof(Boolean),
    offset(incorporate), XtRImmediate, (XtPointer) True},
{
    "install", "Install", XtRBoolean, sizeof(Boolean),
    offset(install), XtRImmediate, (XtPointer) False},
};

static XrmOptionDescRec options[] = {
{
    "-limit",
    "*limit",
    XrmoptionSepArg,
    NULL},
{
    "-inc",
    "*incorporate",
    XrmoptionNoArg,
    "True"},
{
    "+inc",
    "*incorporate",
    XrmoptionNoArg,
    "False"},
{
    "-install",
    "*install",
    XrmoptionNoArg,
    "True"},
{
    "+install",
    "*install",
    XrmoptionNoArg,
    "False"},
{
    "-scale",
    "*ScaleFactor",
    XrmoptionSepArg,
    NULL},
{
    "-ar",
    "*aspectRatio",
    XrmoptionSepArg,
    NULL},
{
    "-a",
    "*autoscale",
    XrmoptionNoArg,
    "True"},
{
    "+a",
    "*autoscale",
    XrmoptionNoArg,
    "False"},
{
    "-p",
    "*proportional",
    XrmoptionNoArg,
    "True"},
{
    "+p",
    "*proportional",
    XrmoptionNoArg,
    "False"}
};


void Quit();
void PopupLabel();
void UpdateLabel();
void PopdownLabel();
void RequestSelection();
void GrabSelection();
static XtActionsRec actions_table[] = {
{"quit", Quit},
{"popup-label", PopupLabel},
{"update-label", UpdateLabel},
{"popdown-label", PopdownLabel},
};

void Quit()
{
    exit(0);
}

void PopupLabel(w, event)
    Widget w;
    XEvent *event;
{
    Position scale_x, scale_y;
    Dimension scale_height;
    Position label_x, label_y;
    Dimension label_height;
    int n;
    Arg wargs[3];

    n = 0;
    XtSetArg(wargs[n], XtNheight, &scale_height); n++;
    XtGetValues(scale, wargs, n);
    XtTranslateCoords(scale, -1, -1, &scale_x, &scale_y);
    
    n = 0;
    XtSetArg(wargs[n], XtNheight, &label_height); n++;
    XtGetValues(shell, wargs, n);
    
    if ((double) event->xbutton.y / (double) scale_height > 0.5) {
	label_x = scale_x;
	label_y = scale_y;
    }
    else {
	label_x = scale_x;
	label_y = scale_y + scale_height - label_height;
    }

    n = 0;
    XtSetArg(wargs[n], XtNx, label_x); n++;
    XtSetArg(wargs[n], XtNy, label_y); n++;
    XtSetValues(shell, wargs, n);
    
    UpdateLabel(w, event);
}

void UpdateLabel(w, event)
    Widget w;
    XEvent *event;
{
    Position x, y;
    Pixel pixel;
    XColor color;
    int n;
    Arg wargs[3];
    char string[80];

    if (SWGetImagePixel(w, event->xbutton.x, event->xbutton.y, &x, &y, &pixel))
	XtPopdown(shell);
    else {
	color.pixel = pixel;
	XQueryColor(XtDisplay(w), colormap, &color);
	sprintf(string, "x:%d y:%d (%04x %04x %04x)", 
		x, y, color.red, color.green, color.blue);
	
	n = 0;
	XtSetArg(wargs[n], XtNlabel, string); n++;    
	XtSetValues(label, wargs, n);
	XtPopup(shell, XtGrabNone);
    }
    
    XFlush(XtDisplay(w));
}

void PopdownLabel(w)
    Widget w;
{
    XtPopdown(shell);
    XFlush(XtDisplay(w));
}

#define NONE 0
#define BITMAP 1
#define GIF 2
#define XWD 3

XImage *ProcessImage();

int main(argc, argv)
    Cardinal argc;
    String   argv[];
{
    Display *dpy;
    int scr;
    Visual *visual;
    String filename = NULL;
    XImage *image;
    XColor *colors, *dcolors;
    Cardinal ncolors = 0, ndcolors, ntcolors;
    int i;
    int type = NONE;
    
    toplevel = XtInitialize(argv[0], "", options, XtNumber(options),
			    &argc, argv);
    dpy = XtDisplay(toplevel);
    scr = DefaultScreen(dpy);
    visual = DefaultVisual(dpy, scr);
    
    XtGetApplicationResources(toplevel, (XtPointer)&app_resources, resources,
			      XtNumber(resources), NULL, 0);
    XtAddActions(actions_table, XtNumber(actions_table));

    while(*++argv && **argv == '-') {
	if (!strcmp(*argv, "-bitmap")) {
	    type = BITMAP;
	    if((filename = *++argv) == NULL)
		Usage();
	}
	else if (!strcmp(*argv, "-gif")) {
	    type = GIF;
	    if((filename = *++argv) == NULL)
		Usage();
	}
	else if (!strcmp(*argv, "-xwd")) {
	    type = XWD;
	    if((filename = *++argv) == NULL)
		Usage();
	}
	else
	    Usage();
    }
    if(filename == NULL || type == NONE)
	Usage();

    switch (type) {
    case BITMAP:
	if (ReadBitmapFile(dpy, scr, filename, &image))
	    XtError("cannot read the file");
	break;
    case GIF:
	if (ReadGifFile(dpy, scr, filename, &image, &colors, &ncolors))
	    XtError("cannot read the file");
	break;
    case XWD:
	if (ReadXwdFile(filename, &image, &colors, &ncolors))
	    XtError("cannot read the file");
	break;
    }

    if (app_resources.limit) ncolors = min(app_resources.limit, ncolors);
    
    scale = XtCreateManagedWidget("scale",
				  scaleWidgetClass, 
				  toplevel, NULL, 0);
    
    shell = XtCreatePopupShell("shell",
			       transientShellWidgetClass,
			       toplevel, NULL, 0);
    
    label = XtCreateManagedWidget("label", 
				  labelWidgetClass,
				  shell, NULL, 0);

    colormap = DefaultColormap(dpy, scr);
    
    if(image->depth > 1) {
	Arg wargs[3];
	int n;
	Pixel foreground;
	Pixel background;
	XColor color;
	
	if (!app_resources.install) {
	    colormap = XCreateColormap(dpy, RootWindow(dpy, scr), 
				       visual, AllocNone);
	    if (app_resources.incorporate) {
		ndcolors = DisplayCells(dpy, scr);
		dcolors = (XColor *) XtMalloc((unsigned) ndcolors *
					      sizeof(XColor));
		for(i = 0; i < ndcolors; ++i)
		{
		    dcolors[i].pixel = i;
		    dcolors[i].flags = DoRed | DoGreen | DoBlue;
		}
		XQueryColors(dpy, DefaultColormap(dpy, scr),
			     dcolors, ndcolors);
		ntcolors = QueryNoTransColors(dpy, scr, visual, colors,
					      ncolors, dcolors, ndcolors);
		for(i = 0; i < ntcolors; ++i)
		    XAllocColor(dpy, colormap, &dcolors[i]);
		XtFree((char*) dcolors);
	    }

	    /* Keep foreground and background colors of label widget */
	    
	    n = 0;
	    XtSetArg(wargs[n], XtNforeground, &foreground); n++;
	    XtSetArg(wargs[n], XtNbackground, &background); n++;
	    XtGetValues(label, wargs, n);
	    
	    color.pixel = foreground;
	    XQueryColor(dpy, DefaultColormap(dpy, scr), &color);
	    XAllocColor(dpy, colormap, &color);
	    foreground = color.pixel;
	    
	    color.pixel = background;
	    XQueryColor(dpy, DefaultColormap(dpy, scr), &color);
	    XAllocColor(dpy, colormap, &color);
	    background = color.pixel;
	    
	    n = 0;
	    XtSetArg(wargs[n], XtNforeground, foreground); n++;
	    XtSetArg(wargs[n], XtNbackground, background); n++;
	    XtSetValues(label, wargs, n);
	}
	image = ProcessImage(dpy, scr, 
			     image, colormap, colors, ncolors, &background);
	n = 0;
	XtSetArg(wargs[n], XtNbackground, background); n++;
	XtSetValues(scale, wargs, n);	
    }
    
    SWSetImage(scale, image);

    XtRealizeWidget(toplevel);
    XtRealizeWidget(shell);
    
    XSetWindowColormap(dpy, XtWindow(toplevel), colormap);
    XSetWindowColormap(dpy, XtWindow(shell), colormap);

    XtMainLoop();
}
    
ReadBitmapFile(dpy, scr, filename, image)
    Display *dpy;
    int scr;
    String filename;
    XImage **image;    
{
    unsigned int width, height;
    int x, y;
    char *data;
    int status;
    
    status = XmuReadBitmapDataFromFile(filename, &width, &height, 
				       &data, &x, &y);
    
    if (status) return status;
    
    *image = XCreateImage(dpy, DefaultVisual(dpy, scr),
			  1, XYBitmap, 0, 
			  data, width, height,
			  8, (width + 7) / 8);
    
    (*image)->byte_order = LSBFirst;
    (*image)->bitmap_bit_order = LSBFirst;
    
    return status;
}

QueryNoTransColors(dpy, screen, visual, colors, ncolors, tcolors, ntcolors)
    Display *dpy;
    int screen;
    Visual *visual;
    XColor *colors;
    int ncolors;
    XColor *tcolors;
    int ntcolors;
{
    Colormap cpm;
    XColor color;
    int nocolors, i, nptcolors;
    
    cpm = XCreateColormap(dpy, RootWindow(dpy, screen), visual, AllocNone);
    
    for(i = 0, nocolors = 0; i < ncolors; ++i) {
	color = colors[i];
	XAllocColor(dpy, cpm, &color);
	if (color.pixel > nocolors) nocolors = color.pixel;
    }
    
    for (nptcolors = 0;
	 nocolors + 1 < ntcolors && nptcolors < ntcolors;
	 nptcolors++) {
	color = tcolors[nptcolors];
	XAllocColor(dpy, cpm, &color);
	if (color.pixel > nocolors) nocolors = color.pixel;
    }
    
    XFreeColormap(dpy, cpm);
    
    return nptcolors;
}










